@@ -1,12 +1,11 @@
 /*
-	htpdate v0.9.1
+	htpdate v1.1.0
 
-	Eddy Vervest <eddy@cleVervest.com>
-	http://www.clevervest.com/htp
+	Eddy Vervest <eddy@vervest.org>
+	http://www.vervest.org/htp
 
 	Synchronize local workstation with time offered by remote web servers
 
-	Extract date/time stamp from web server response
 	This program works with the timestamps return by web servers,
 	formatted as specified by HTTP/1.1 (RFC 2616, RFC 1123).
 
@@ -21,7 +20,7 @@
 
 	~# htpdate -a www.linux.org www.freebsd.org
 
-	...see manpage for more parameters
+	...see man page for more details
 
 
 	This program is free software; you can redistribute it and/or
@@ -44,36 +43,36 @@
 #include <netinet/in.h>
 #include <sys/socket.h>
 #include <netdb.h>
-#include <sys/time.h>
 #include <time.h>
+#include <sys/time.h>
+#include <sys/timex.h>
 #include <syslog.h>
 #include <stdarg.h>
 #include <limits.h>
 #include <pwd.h>
 #include <grp.h>
 
-#define VERSION 				"0.9.1"
+#define VERSION 				"1.1.0"
 #define	MAX_HTTP_HOSTS			15				/* 16 web servers */
 #define	DEFAULT_HTTP_PORT		"80"
 #define	DEFAULT_PROXY_PORT		"8080"
 #define	DEFAULT_IP_VERSION		PF_UNSPEC		/* IPv6 and IPv4 */
 #define	DEFAULT_HTTP_VERSION	"1"				/* HTTP/1.1 */
 #define	DEFAULT_TIME_LIMIT		31536000		/* 1 year */
-#define	DEFAULT_MIN_SLEEP		10				/* 2^10 => 17 minutes */
-#define	DEFAULT_MAX_SLEEP		18				/* 2^18 => 72 hours */
-#define	RETRY					2				/* Poll attempts */
+#define	DEFAULT_MIN_SLEEP		1800			/* 30 minutes */
+#define	DEFAULT_MAX_SLEEP		115200			/* 32 hours */
+#define	MAX_DRIFT				32768000		/* 500 PPM */
+#define	MAX_ATTEMPT				2				/* Poll attempts */
 #define	DEFAULT_PID_FILE		"/var/run/htpdate.pid"
-#define	BUFFERSIZE				2048
+#define	URLSIZE					128
+#define	BUFFERSIZE				1024
 
-#define SIGN(x) (x < 0 ? (-1) : 1)
+#define sign(x) (x < 0 ? (-1) : 1)
 
 
-/* By default we turn off "debug", "log" and "daemonize" mode  */
-#ifdef DEBUG
+/* By default we turn off "debug" and "log" mode  */
 static int		debug = 0;
-#endif
 static int		logmode = 0;
-static int		daemonize = 0;
 static time_t	gmtoffset;
 
 
@@ -129,10 +128,10 @@ static void splithostport( char **host, char **port ) {
 /* Printlog is a slighty modified version from the one used in rdate */
 static void printlog( int is_error, char *format, ... ) {
 	va_list args;
-	int n;
 	char buf[128];
+
 	va_start(args, format);
-	n = vsnprintf(buf, sizeof(buf), format, args);
+	(void) vsnprintf(buf, sizeof(buf), format, args);
 	va_end(args);
 
     if ( logmode )
@@ -142,19 +141,18 @@ static void printlog( int is_error, char *format, ... ) {
 }
 
 
-static long getHTTPdate( char *host, char *port, char *proxy, char *proxyport, char *httpversion, int ipversion, unsigned long when ) {
+static long getHTTPdate( char *host, char *port, char *proxy, char *proxyport, char *httpversion, int ipversion, int when ) {
 	int					server_s;
 	int					rc;
 	struct addrinfo		hints, *res, *res0;
 	struct tm			tm;
-	struct timeval		timevalue = {.tv_sec = LONG_MAX};
-	struct timeval		timeofday = {.tv_sec = 0};
-#ifdef DEBUG
+	struct timeval		timevalue = {LONG_MAX, 0};
+	struct timeval		timeofday;
+	struct timespec		sleepspec, remainder;
 	long				rtt;
-#endif
 	char				buffer[BUFFERSIZE];
-	char				remote_time[25] = {};
-	char				url[128] = {};
+	char				remote_time[25] = { '\0' };
+	char				url[URLSIZE] = { '\0' };
 	char				*pdate = NULL;
 
 
@@ -176,14 +174,14 @@ static long getHTTPdate( char *host, char *port, char *proxy, char *proxyport, c
 	if ( proxy == NULL ) {
 		rc = getaddrinfo( host, port, &hints, &res0 );
 	} else {
-		sprintf( url, "http://%s:%s", host, port);
+		snprintf( url, URLSIZE, "http://%s:%s", host, port);
 		rc = getaddrinfo( proxy, proxyport, &hints, &res0 );
 	}
 
 	/* Was the hostname and service resolvable? */
 	if ( rc ) {
 		printlog( 1, "%s host or service unavailable", host );
-		return( LONG_MAX );
+		return(0);				/* Assume correct time */
 	}
 
 	/* Build a combined HTTP/1.0 and 1.1 HEAD request
@@ -192,7 +190,7 @@ static long getHTTPdate( char *host, char *port, char *proxy, char *proxyport, c
 	   Connection: close, allows the server the immediately close the
 	   connection after sending the response.
 	*/
-	sprintf(buffer, "HEAD %s/ HTTP/1.%s\r\nHost: %s\r\nUser-Agent: htpdate/%s\r\nPragma: no-cache\r\nCache-Control: no-cache\r\nConnection: close\r\n\r\n", url, httpversion, host, VERSION);
+	snprintf(buffer, BUFFERSIZE, "HEAD %s/ HTTP/1.%s\r\nHost: %s\r\nUser-Agent: htpdate/"VERSION"\r\nPragma: no-cache\r\nCache-Control: no-cache\r\nConnection: close\r\n\r\n", url, httpversion, host);
 
 	/* Loop through the available canonical names */
 	res = res0;
@@ -212,27 +210,28 @@ static long getHTTPdate( char *host, char *port, char *proxy, char *proxyport, c
 		break;
 	} while ( ( res = res->ai_next ) );
 
+	freeaddrinfo(res0);
+
 	if ( rc ) {
 		printlog( 1, "%s connection failed", host );
-		return( LONG_MAX );
+		return(0);				/* Assume correct time */
 	}
 
-	/* Wait till we reach the desired time, "when" */
+	/* Initialize timer */
 	gettimeofday(&timeofday, NULL);
 
-#ifdef DEBUG
 	/* Initialize RTT (start of measurement) */
 	rtt = timeofday.tv_sec;
-#endif
 
+	/* Wait till we reach the desired time, "when" */
+	sleepspec.tv_sec = 0;
 	if ( when >= timeofday.tv_usec ) {
-		usleep( when - timeofday.tv_usec );
+		sleepspec.tv_nsec = ( when - timeofday.tv_usec ) * 1000;
 	} else {
-		usleep( 1000000 + when - timeofday.tv_usec );
-#ifdef DEBUG
+		sleepspec.tv_nsec = ( 1000000 + when - timeofday.tv_usec ) * 1000;
 		rtt++;
-#endif
 	}
+	nanosleep( &sleepspec, &remainder );
 
 	/* Send HEAD request */
 	if ( send(server_s, buffer, strlen(buffer), 0) < 0 )
@@ -257,11 +256,9 @@ static long getHTTPdate( char *host, char *port, char *proxy, char *proxyport, c
 
 		gettimeofday(&timeofday, NULL);
 
-#ifdef DEBUG
 		/* rtt contains round trip time in micro seconds, now! */
 		rtt = ( timeofday.tv_sec - rtt ) * 1000000 + \
 			timeofday.tv_usec - when;
-#endif
 
 		/* Look for the line that contains Date: */
 		if ( (pdate = strstr(buffer, "Date: ")) != NULL ) {
@@ -275,13 +272,11 @@ static long getHTTPdate( char *host, char *port, char *proxy, char *proxyport, c
 				printlog( 1, "%s unknown time format", host );
 			}
 
-#ifdef DEBUG
 			/* Print host, raw timestamp, round trip time */
 			if ( debug )
 				printlog( 0, "%-25s %s (%.3f) => %li", host, remote_time, \
 				  rtt * 1e-6, timevalue.tv_sec - timeofday.tv_sec \
 				  + gmtoffset );
-#endif
 
 		} else {
 			printlog( 1, "%s no timestamp", host );
@@ -309,34 +304,46 @@ static int setclock( double timedelta, int setmode ) {
 
 	switch ( setmode ) {
 
-	case 0:						/* no time adjustment, just print time */
+	case 0:						/* No time adjustment, just print time */
 		printlog( 0, "Offset %.3f seconds", timedelta );
 		return(0);
 
-	case 1:						/* adjust time smoothly */
+	case 1:						/* Adjust time smoothly */
 		timeofday.tv_sec  = (long)timedelta;	
 		timeofday.tv_usec = (long)((timedelta - timeofday.tv_sec) * 1000000);	
 
 		printlog( 0, "Adjusting %.3f seconds", timedelta );
 
-		/* become root */
-		seteuid(0);
-		return( adjtime(&timeofday, NULL) );
+		/* Become root */
+		if ( seteuid(0) ) {
+			printlog( 1, "seteuid()" );
+			exit(1);
+		} else {
+			return( adjtime(&timeofday, NULL) );
+		}
 
-	case 2:						/* set time */
+	case 2:					/* Set time */
 		printlog( 0, "Setting %.3f seconds", timedelta );
 
 		gettimeofday( &timeofday, NULL );
 		timedelta += ( timeofday.tv_sec + timeofday.tv_usec*1e-6 );
 
 		timeofday.tv_sec  = (long)timedelta;	
-		timeofday.tv_usec = (long)((timedelta - timeofday.tv_sec) * 1000000);	
+		timeofday.tv_usec = (long)(timedelta - timeofday.tv_sec) * 1000000;	
 
 		printlog( 0, "Set: %s", asctime(localtime(&timeofday.tv_sec)) );
 
-		/* become root */
-		seteuid(0);
-		return( settimeofday(&timeofday, NULL) );
+		/* Become root */
+		if ( seteuid(0) ) {
+			printlog( 1, "seteuid()" );
+			exit(1);
+		} else {
+			return( settimeofday(&timeofday, NULL) );
+		}
+
+	case 3:					/* Set frequency, but first an adjust */
+		return( setclock( timedelta, 1 ) );
+
 
 	default:
 		return(-1);
@@ -346,16 +353,46 @@ static int setclock( double timedelta, int setmode ) {
 }
 
 
-/* Display help page */
+static int htpdate_adjtimex( double drift ) {
+	struct timex		tmx;
+	long				freq;
+
+	/* Read current kernel frequency */
+	tmx.modes = 0;
+	ntp_adjtime(&tmx);
+
+	/* Calculate new frequency */
+	freq = (long)(65536e6 * drift);
+
+	/* Take the average of current and new drift values */
+	tmx.freq = tmx.freq + (freq >> 1);
+	if ( (tmx.freq < -MAX_DRIFT) || (tmx.freq > MAX_DRIFT) )
+		tmx.freq = sign(tmx.freq) * MAX_DRIFT;
+
+	printlog( 0, "Adjusting frequency %li", tmx.freq );
+	tmx.modes = MOD_FREQUENCY;
+
+	/* Become root */
+	if ( seteuid(0) ) {
+		printlog( 1, "seteuid()" );
+		exit(1);
+	} else {
+		return( ntp_adjtime(&tmx) );
+	}
+
+}
+
+
 static void showhelp() {
-	printf("htpdate version %s\n\
-Usage: htpdate [-046adhlqstD] [-i pid file] [-m minpoll] [-M maxpoll]\n\
+	puts("htpdate version "VERSION"\n\
+Usage: htpdate [-046abdhlqstxD] [-i pid file] [-m minpoll] [-M maxpoll]\n\
          [-p precision] [-P <proxyserver>[:port]] [-u user[:group]]\n\
          <host[:port]> ...\n\n\
   -0    HTTP/1.0 request\n\
   -4    Force IPv4 name resolution only\n\
   -6    Force IPv6 name resolution only\n\
   -a    adjust time smoothly\n\
+  -b    burst mode\n\
   -d    debug mode\n\
   -D    daemon mode\n\
   -h    help\n\
@@ -369,8 +406,9 @@ Usage: htpdate [-046adhlqstD] [-i pid file] [-m minpoll] [-M maxpoll]\n\
   -s    set time\n\
   -t    turn off sanity time check\n\
   -u    run daemon as user\n\
+  -x    adjust kernel clock\n\
   host  web server hostname or ip address (maximum of 16)\n\
-  port  port number (default 80 and 8080 for proxy server)\n\n", VERSION);
+  port  port number (default 80 and 8080 for proxy server)\n");
 
 	return;
 }
@@ -384,13 +422,13 @@ static void runasdaemon( char *pidfile ) {
 	/* Check if htpdate is already running (pid exists)*/
 	pid_file = fopen(pidfile, "r");
 	if ( pid_file ) {
-		fprintf( stderr, "htpdate already running" );
+		fputs( "htpdate already running\n", stderr );
 		exit(1);
 	}
 
 	pid = fork();
 	if ( pid < 0 ) {
-		fprintf( stderr, "fork()" );
+		fputs( "fork()\n", stderr );
 		exit(1);
 	}
 
@@ -413,8 +451,10 @@ static void runasdaemon( char *pidfile ) {
 	umask(0);
 
 	/* Change the current working directory */
-	if ( chdir("/") < 0 )
+	if ( chdir("/") < 0 ) {
+		printlog( 1, "chdir()" );
 		exit(1);
+	}
 
 	/* Second fork, to become the grandchild */
 	pid = fork();
@@ -441,17 +481,19 @@ static void runasdaemon( char *pidfile ) {
 
 
 int main( int argc, char *argv[] ) {
-	char				*host = NULL, *proxy = NULL;
-	char				*port, *proxyport;
+	char				*host = NULL, *proxy = NULL, *proxyport = NULL;
+	char				*port;
 	char				*httpversion = DEFAULT_HTTP_VERSION;
 	char				*pidfile = DEFAULT_PID_FILE;
 	char				*user = NULL, *userstr = NULL, *group = NULL;
-	double				timeavg = 0, sumdelta = 0, drift;
-	int					timedelta[MAX_HTTP_HOSTS], timestamp;
-	int                 sumtimes, numservers, validtimes, goodtimes, mean;
+	long long			sumtimes;
+	double				timeavg, drift = 0;
+	int					timedelta[(MAX_HTTP_HOSTS+1)*(MAX_HTTP_HOSTS+1)-1], timestamp;
+	int                 numservers, validtimes, goodtimes, mean;
 	int					nap = 0, when = 500000, precision = 0;
-	int					setmode = 0, try, offsetdetect;
-	int					i, param;
+	int					setmode = 0, burstmode = 0, try, offsetdetect;
+	int					i, burst, param;
+	int					daemonize = 0;
 	int					ipversion = DEFAULT_IP_VERSION;
 	int					timelimit = DEFAULT_TIME_LIMIT;
 	int					minsleep = DEFAULT_MIN_SLEEP;
@@ -468,7 +510,7 @@ int main( int argc, char *argv[] ) {
 
 
 	/* Parse the command line switches and arguments */
-	while ( (param = getopt(argc, argv, "046adhi:lm:p:qstu:DM:P:") ) != -1)
+	while ( (param = getopt(argc, argv, "046abdhi:lm:p:qstu:xDM:P:") ) != -1)
 	switch( param ) {
 
 		case '0':			/* HTTP/1.0 */
@@ -483,23 +525,24 @@ int main( int argc, char *argv[] ) {
 		case 'a':			/* adjust time */
 			setmode = 1;
 			break;
+		case 'b':			/* burst mode */
+			burstmode = 1;
+			break;
 		case 'd':			/* turn debug on */
-#ifdef DEBUG
 			debug = 1;
-#endif
 			break;
 		case 'h':			/* show help */
 			showhelp();
 			exit(0);
-		case 'i':			/* pid file help */
+		case 'i':			/* pid file */
 			pidfile = (char *)optarg;
 			break;
-		case 'l':			/* pid file help */
+		case 'l':			/* log mode */
 			logmode = 1;
 			break;
 		case 'm':			/* minimum poll interval */
 			if ( ( minsleep = atoi(optarg) ) <= 0 ) {
-				fprintf( stderr, "Invalid sleep time" );
+				fputs( "Invalid sleep time\n", stderr );
 				exit(1);
 			}
 			sleeptime = minsleep;
@@ -507,7 +550,7 @@ int main( int argc, char *argv[] ) {
 		case 'p':			/* precision */
 			precision = atoi(optarg) ;
 			if ( (precision <= 0) || (precision >= 500) ) {
-				fprintf( stderr, "Invalid precision" );
+				fputs( "Invalid precision\n", stderr );
 				exit(1);
 			}
 			precision *= 1000;
@@ -543,13 +586,16 @@ int main( int argc, char *argv[] ) {
 				}
 			}
 			break;
+		case 'x':			/* adjust time and "kernel" */
+			setmode = 3;
+			break;
 		case 'D':			/* run as daemon */
 			daemonize = 1;
 			logmode = 1;
 			break;
 		case 'M':			/* maximum poll interval */
 			if ( ( maxsleep = atoi(optarg) ) <= 0 ) {
-				fprintf( stderr, "Invalid sleep time" );
+				fputs( "Invalid sleep time\n", stderr );
 				exit(1);
 			}
 			break;
@@ -573,13 +619,13 @@ int main( int argc, char *argv[] ) {
 	/* Exit if too many servers are specified */
 	numservers = argc - optind;
 	if ( numservers > MAX_HTTP_HOSTS ) {
-		fprintf( stderr, "Too many servers" );
+		fputs( "Too many servers\n", stderr );
 		exit(1);
 	}
 
 	/* One must be "root" to change the system time */
 	if ( (getuid() != 0) && (setmode || daemonize) ) {
-		fprintf( stderr, "Only root can change time\n" );
+		fputs( "Only root can change time\n", stderr );
 		exit(1);
 	}
 
@@ -589,8 +635,6 @@ int main( int argc, char *argv[] ) {
 		/* Query only mode doesn't exist in daemon mode */
 		if ( !setmode )
 			setmode = 1;
-	} else {
-		precision = 0;
 	}
 
 	/* Now we are root, we drop the privileges (if specified) */
@@ -602,12 +646,17 @@ int main( int argc, char *argv[] ) {
     gmtoffset -= mktime(gmtime(&gmtoffset));
 
 	/* In case we have more than one web server defined, we
-	   spread the polls equal within a second and take a "nap" in between.
+	   spread the polls equal within a second and take a "nap" in between
 	*/
 	if ( numservers > 1 )
-		nap = 1000000 / (numservers + 1);
-	else
+		if ( precision && (numservers > 2) )
+			nap = (1000000 - 2*precision) / (numservers - 1);
+		else
+			nap = 1000000 / (numservers + 1);
+	else {
 		precision = 0;
+		nap = 500000;
+	}
 
 	/* Infinite poll cycle loop in daemonize mode */
 	do {
@@ -629,51 +678,55 @@ int main( int argc, char *argv[] ) {
 		port = DEFAULT_HTTP_PORT;
 		splithostport( &host, &port );
 
-		/* Retry if first poll shows time offset */
-		try = RETRY;
-		do {
-			timestamp = getHTTPdate( host, port, proxy, proxyport,\
-					httpversion, ipversion, when );
-			try--;
-		} while ( timestamp && try && daemonize );
-
-		/* Only include valid responses in timedelta[] */
-		if ( timestamp < timelimit && timestamp > -timelimit ) {
-			timedelta[validtimes] = timestamp;
-			validtimes++;
+		/* if burst mode, reset "when" */
+		if ( burstmode ) {
+			if ( precision )
+				when = precision;
+			else
+				when = nap;
 		}
 
-		/* If we detected a time offset, set the flag */
-		if ( timestamp )
+		burst = 0;
+		do {
+			/* Retry if first poll shows time offset */
+			try = MAX_ATTEMPT;
+			do {
+				if ( debug ) printlog( 0, "burst: %d try: %d when: %d", \
+					burst + 1, MAX_ATTEMPT - try + 1, when );
+				timestamp = getHTTPdate( host, port, proxy, proxyport,\
+						httpversion, ipversion, when );
+				try--;
+			} while ( timestamp && try );
+
+			/* Only include valid responses in timedelta[] */
+			if ( timestamp < timelimit && timestamp > -timelimit ) {
+				timedelta[validtimes] = timestamp;
+				validtimes++;
+			}
+
+			/* If we detected a time offset, set the flag */
+			if ( timestamp )
 				offsetdetect = 1;
 
+			/* Take a nap, to spread polls equally within a second.
+			   Example:
+			   2 servers => 0.333, 0.666
+			   3 servers => 0.250, 0.500, 0.750
+			   4 servers => 0.200, 0.400, 0.600, 0.800
+			   ...
+			   nap = 1000000 / (#servers + 1)
+
+			   or when "precision" is specified, a different algorithm is used
+			*/
+			when += nap;
+
+			burst++;
+		} while ( burst < (argc - optind) * burstmode );
+
 		/* Sleep for a while, unless we detected a time offset */
 		if ( daemonize && !offsetdetect )
-			sleep( (1 << sleeptime) / numservers );
+			sleep( sleeptime / numservers );
 
-		/* Take a nap, to spread polls equally within a second.
-		   Example:
-		   2 servers => 0.333, 0.666
-		   3 servers => 0.250, 0.500, 0.750
-		   4 servers => 0.200, 0.400, 0.600, 0.800
-		   ...
-		   nap = 1000000 / (#servers + 1)
-
-		   or when "precision" is specified, use a different algorithm
-		   Example with a precision of 200000:
-		   2 servers => 0.200, 0.800
-		   3 servers => 0.200, 0.800, 0.200
-		   4 servers => 0.200, 0.800, 0.200, 0.800
-		   ...
-		*/
-		if ( precision ) {
-			if ( when > 500000 )
-				when = precision;
-			else
-				when = 1000000 - precision;
-		} else {
-			when += nap;
-		}
 	}
 
 	/* Sort the timedelta results */
@@ -698,13 +751,11 @@ int main( int argc, char *argv[] ) {
 
 		timeavg = sumtimes/(double)goodtimes;
 
-#ifdef DEBUG
 		if ( debug ) {
-			printlog( 0, "#: %d, mean: %d, average: %.3f", goodtimes, \
+			printlog( 0, "#: %d mean: %d average: %.3f", goodtimes, \
 					mean, timeavg );
-			printlog( 0, "Timezone: GMT%+li (%s,%s)", gmtoffset / 3600, tzname[0], tzname[1] );
+			printlog( 0, "Timezone: GMT%+li (%s,%s)", gmtoffset/3600, tzname[0], tzname[1] );
 		}
-#endif
 
 		/* Do I really need to change the time?  */
 		if ( sumtimes || !daemonize ) {
@@ -712,56 +763,65 @@ int main( int argc, char *argv[] ) {
 			   (< +-1 second), adjust the time with the value of precision
 			*/
 			if ( precision && sumtimes < goodtimes && sumtimes > -goodtimes )
-				timeavg = (double)precision / 1000000 * SIGN(sumtimes);
+				timeavg = (double)precision / 1000000 * sign(sumtimes);
 
-			/* Correct the clock in daemon mode */
-			if ( setclock( timeavg, setmode ) )
-				printlog( 1, "Time change failed" );
+			/* Correct the clock, if not in "adjtimex" mode */
+			if ( setclock( timeavg, setmode ) < 0 )
+					printlog( 1, "Time change failed" );
 
 			/* Drop root privileges again */
 			if ( sw_uid ) seteuid( sw_uid );
-			if ( sw_gid ) setegid( sw_gid );
 
 			if ( daemonize ) {
 				if ( starttime ) {
 					/* Calculate systematic clock drift */
-					sumdelta += timeavg;
-					drift = sumdelta * 86400 / (time(NULL) - starttime);
-					printlog( 0, "Drift %.2f PPM, %.2f sec/day", \
-							drift / 0.0864, drift );
+					drift = timeavg / ( time(NULL) - starttime );
+					printlog( 0, "Drift %.2f PPM, %.2f s/day", \
+							drift*1e6, drift*86400 );
+
+					/* Adjust system clock */
+					if ( setmode == 3 ) {
+						starttime = time(NULL);
+						/* Adjust the kernel clock */
+						if ( htpdate_adjtimex( drift ) < 0 )
+							printlog( 1, "Frequency change failed" );
+
+						/* Drop root privileges again */
+						if ( sw_uid ) seteuid( sw_uid );
+					}
 				} else {
 					starttime = time(NULL);
 				}
 
-				/* Decrease polling interval */
-				if ( sleeptime > minsleep )
-					sleeptime--;
+				/* Decrease polling interval to minimum */
+				sleeptime = minsleep;
 
-				/* Sleep for minsleep, after a time adjust or set */
-				sleep( 1 << minsleep );
+				/* Sleep for 30 minutes after a time adjust or set */
+				sleep( DEFAULT_MIN_SLEEP );
 			}
 		} else {
 			/* Increase polling interval */
 			if ( sleeptime < maxsleep )
-				sleeptime++;
+				sleeptime <<= 1;
 		}
 
 	} else {
 		printlog( 1, "No server suitable for synchronization found" );
 		/* Sleep for minsleep to avoid flooding */
 		if ( daemonize )
-			sleep( 1 << minsleep );
+			sleep( minsleep );
 		else
 			exit(1);
 	}
 
 	/* After first poll cycle do not step through time, only adjust */
-	setmode = 1;
+	if ( setmode != 3 ) {
+		setmode = 1;
+	}
 
 	} while ( daemonize );		/* end of infinite while loop */
 
 	exit(0);
 }
 
-
-/* vi:set ts=4: */
+/* vim: set ts=4 sw=4: */
